from typing import Dict

import numpy as np
from jina.executors.crafters import BaseCrafter


class ArrayStringReader(BaseCrafter):
    """
    Convert string of numbers into a numpy array and save to the Document.

    Numbers are split on the provided delimiter, default is comma (,)

    :param delimiter: delimiter between numbers
    :param as_type: The numpy array will be this type
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments
    """

    def __init__(self, delimiter: str = ',', as_type: str = 'float32', *args, **kwargs):
        """Set constructor."""
        super().__init__(*args, **kwargs)
        self.delimiter = delimiter
        self.as_type = as_type

    def craft(self, text: str, *args, **kwargs) -> Dict:
        """
        Split string into numbers and convert to numpy array.

        :param text: the raw text to be converted
        :return: a dict with the created numpy array
        :param args:  Additional positional arguments
        :param kwargs: Additional keyword arguments
        """
        _string = text.split(self.delimiter)
        _array = np.array(_string)

        try:
            _array = _array.astype(self.as_type)
        except TypeError:
            self.logger.error(
                f'Data type {self.as_type} is not understood. '
                f'Please refer to the list of data types from Numpy.')
        except ValueError:
            self.logger.error(
                f'Data type mismatch. Cannot convert input to {self.as_type}.')

        return dict(blob=_array)



"""Argparser module for hub build"""
from ..helper import add_arg_group
from ...enums import BuildTestLevel


def mixin_hub_build_parser(parser):
    """Add the arguments for hub build to the parser

    :param parser: the parser configure
    """
    gp = add_arg_group(parser, title='Build')
    gp.add_argument('path', type=str, help='path to the directory containing '
                                           'Dockerfile, manifest.yml, README.md '
                                           'zero or more yaml config, '
                                           'zero or more Python file. '
                                           'All files in this directory will be shipped into a Docker image')
    gp.add_argument('-f', '--file', type=str, default='Dockerfile',
                    help='Name of the Dockerfile (Default is `path/Dockerfile`')
    gp.add_argument('--pull', action='store_true', default=False,
                    help='If set, downloads any updates to the FROM image in Dockerfiles')
    gp.add_argument('--push', action='store_true', default=False,
                    help='If set, push the built image to the registry')
    gp.add_argument('--dry-run', action='store_true', default=False,
                    help='If set, only check path and validity, no real building')
    gp.add_argument('--prune-images', action='store_true', default=False,
                    help='If set, prune unused images after building, this often saves disk space')
    gp.add_argument('--raise-error', action='store_true', default=False,
                    help='If set, raise any error and exit with code 1')
    gp.add_argument('--test-uses', action='store_true', default=False,
                    help='If set, after the build, test the image in `--uses` with different level')
    gp.add_argument('--test-level', type=BuildTestLevel.from_string,
                    choices=list(BuildTestLevel), default=BuildTestLevel.FLOW,
                    help='If set, the test level when `--test-uses` is set, `NONE` means no test')
    gp.add_argument('--timeout-ready', type=int, default=60000,
                    help='The timeout in millisecond to give for the Pod to start before considering a test failed')
    gp.add_argument('--host-info', action='store_true', default=False,
                    help='If set, store the host information during bookkeeping')
    gp.add_argument('--daemon', action='store_true', default=False,
                    help='If set, run the test Pea/Pod as a daemon process, see `jina pea --help` for details')
    gp.add_argument('--no-overwrite', action='store_true', default=False,
                    help='If set, do not overwrite existing images (based on module version and jina version)')



import warnings
from typing import Union, Iterable, TextIO, Dict

import numpy as np

from ...clients.base import InputType, CallbackFnType
from ...enums import DataInputType
from ...helper import deprecated_alias


class CRUDFlowMixin:
    """The synchronous version of the Mixin for CRUD in Flow"""

    @deprecated_alias(input_fn=('inputs', 0))
    def train(self, inputs: InputType,
              on_done: CallbackFnType = None,
              on_error: CallbackFnType = None,
              on_always: CallbackFnType = None,
              **kwargs):
        """Do training on the current Flow

        :param inputs: An iterator of bytes. If not given, then you have to specify it in **kwargs**.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :return: results
        """
        warnings.warn(f'{self.train} is under heavy refactoring', FutureWarning)
        return self._get_client(**kwargs).train(inputs, on_done, on_error, on_always, **kwargs)

    @deprecated_alias(input_fn=('inputs', 0), buffer=('inputs', 1), callback=('on_done', 1), output_fn=('on_done', 1))
    def index_ndarray(self, array: 'np.ndarray',
                      axis: int = 0,
                      size: int = None,
                      shuffle: bool = False,
                      on_done: CallbackFnType = None,
                      on_error: CallbackFnType = None,
                      on_always: CallbackFnType = None,
                      **kwargs):
        """Using numpy ndarray as the index source for the current Flow

        :param array: the numpy ndarray data source
        :param axis: iterate over that axis
        :param size: the maximum number of the sub arrays
        :param shuffle: shuffle the the numpy data source beforehand
        :param on_done: the callback function to invoke after indexing
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :return: results
        """
        from ...clients.sugary_io import _input_ndarray
        return self._get_client(**kwargs).index(_input_ndarray(array, axis, size, shuffle),
                                                on_done, on_error, on_always, data_type=DataInputType.CONTENT, **kwargs)

    @deprecated_alias(input_fn=('inputs', 0), buffer=('inputs', 1), callback=('on_done', 1), output_fn=('on_done', 1))
    def search_ndarray(self, array: 'np.ndarray', axis: int = 0, size: int = None, shuffle: bool = False,
                       on_done: CallbackFnType = None,
                       on_error: CallbackFnType = None,
                       on_always: CallbackFnType = None,
                       **kwargs):
        """Use a numpy ndarray as the query source for searching on the current Flow

        :param array: the numpy ndarray data source
        :param axis: iterate over that axis
        :param size: the maximum number of the sub arrays
        :param shuffle: shuffle the the numpy data source beforehand
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        """
        from ...clients.sugary_io import _input_ndarray
        self._get_client(**kwargs).search(_input_ndarray(array, axis, size, shuffle),
                                          on_done, on_error, on_always, data_type=DataInputType.CONTENT, **kwargs)

    @deprecated_alias(input_fn=('inputs', 0), buffer=('inputs', 1), callback=('on_done', 1), output_fn=('on_done', 1))
    def index_lines(self,
                    lines: Union[Iterable[str], TextIO] = None,
                    filepath: str = None,
                    size: int = None,
                    sampling_rate: float = None,
                    read_mode: str = 'r',
                    line_format: str = 'json',
                    field_resolver: Dict[str, str] = None,
                    on_done: CallbackFnType = None,
                    on_error: CallbackFnType = None,
                    on_always: CallbackFnType = None,
                    **kwargs):
        """ Use a list of lines as the index source for indexing on the current Flow
        :param lines: a list of strings, each is considered as d document
        :param filepath: a text file that each line contains a document
        :param size: the maximum number of the documents
        :param sampling_rate: the sampling rate between [0, 1]
        :param read_mode: specifies the mode in which the file
                is opened. 'r' for reading in text mode, 'rb' for reading in binary
        :param line_format: the format of each line: ``json`` or ``csv``
        :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
            names defined in Protobuf. This is only used when the given ``document`` is
            a JSON string or a Python dict.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :return: results
        """
        from ...clients.sugary_io import _input_lines
        return self._get_client(**kwargs).index(
            _input_lines(lines, filepath,
                         size=size,
                         sampling_rate=sampling_rate,
                         read_mode=read_mode,
                         line_format=line_format,
                         field_resolver=field_resolver),
            on_done, on_error, on_always, data_type=DataInputType.AUTO, **kwargs)

    def index_ndjson(self,
                     lines: Union[Iterable[str], TextIO],
                     field_resolver: Dict[str, str] = None,
                     size: int = None,
                     sampling_rate: float = None,
                     on_done: CallbackFnType = None,
                     on_error: CallbackFnType = None,
                     on_always: CallbackFnType = None,
                     **kwargs):
        """ Use a list of lines as the index source for indexing on the current Flow
        :param lines: a list of strings, each is considered as d document
        :param size: the maximum number of the documents
        :param sampling_rate: the sampling rate between [0, 1]
        :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
            names defined in Protobuf. This is only used when the given ``document`` is
            a JSON string or a Python dict.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :return: results
        """
        from ...clients.sugary_io import _input_ndjson
        return self._get_client(**kwargs).index(
            _input_ndjson(lines,
                          size=size,
                          sampling_rate=sampling_rate,
                          field_resolver=field_resolver),
            on_done, on_error, on_always, data_type=DataInputType.AUTO, **kwargs)

    def index_csv(self,
                  lines: Union[Iterable[str], TextIO],
                  field_resolver: Dict[str, str] = None,
                  size: int = None,
                  sampling_rate: float = None,
                  on_done: CallbackFnType = None,
                  on_error: CallbackFnType = None,
                  on_always: CallbackFnType = None,
                  **kwargs):
        """ Use a list of lines as the index source for indexing on the current Flow
        :param lines: a list of strings, each is considered as d document
        :param size: the maximum number of the documents
        :param sampling_rate: the sampling rate between [0, 1]
        :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
            names defined in Protobuf. This is only used when the given ``document`` is
            a JSON string or a Python dict.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :return: results
        """
        from ...clients.sugary_io import _input_csv
        return self._get_client(**kwargs).index(
            _input_csv(lines,
                       size=size,
                       sampling_rate=sampling_rate,
                       field_resolver=field_resolver),
            on_done, on_error, on_always, data_type=DataInputType.AUTO, **kwargs)

    def search_csv(self,
                   lines: Union[Iterable[str], TextIO],
                   field_resolver: Dict[str, str] = None,
                   size: int = None,
                   sampling_rate: float = None,
                   on_done: CallbackFnType = None,
                   on_error: CallbackFnType = None,
                   on_always: CallbackFnType = None,
                   **kwargs):
        """ Use a list of lines as the index source for indexing on the current Flow
        :param lines: a list of strings, each is considered as d document
        :param size: the maximum number of the documents
        :param sampling_rate: the sampling rate between [0, 1]
        :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
            names defined in Protobuf. This is only used when the given ``document`` is
            a JSON string or a Python dict.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :return: results
        """
        from ...clients.sugary_io import _input_csv
        return self._get_client(**kwargs).search(
            _input_csv(lines,
                       size=size,
                       sampling_rate=sampling_rate,
                       field_resolver=field_resolver),
            on_done, on_error, on_always, data_type=DataInputType.AUTO, **kwargs)

    @deprecated_alias(input_fn=('inputs', 0), buffer=('inputs', 1), callback=('on_done', 1), output_fn=('on_done', 1))
    def index_files(self, patterns: Union[str, Iterable[str]], recursive: bool = True,
                    size: int = None, sampling_rate: float = None, read_mode: str = None,
                    on_done: CallbackFnType = None,
                    on_error: CallbackFnType = None,
                    on_always: CallbackFnType = None,
                    **kwargs):
        """ Use a set of files as the index source for indexing on the current Flow
        :param patterns: The pattern may contain simple shell-style wildcards, e.g. '\*.py', '[\*.zip, \*.gz]'
        :param recursive: If recursive is true, the pattern '**' will match any files and
                    zero or more directories and subdirectories.
        :param size: the maximum number of the files
        :param sampling_rate: the sampling rate between [0, 1]
        :param read_mode: specifies the mode in which the file
                is opened. 'r' for reading in text mode, 'rb' for reading in binary mode
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :return: results
        """
        from ...clients.sugary_io import _input_files
        return self._get_client(**kwargs).index(_input_files(patterns, recursive, size, sampling_rate, read_mode),
                                                on_done, on_error, on_always, data_type=DataInputType.CONTENT, **kwargs)

    @deprecated_alias(input_fn=('inputs', 0), buffer=('inputs', 1), callback=('on_done', 1), output_fn=('on_done', 1))
    def search_files(self, patterns: Union[str, Iterable[str]], recursive: bool = True,
                     size: int = None, sampling_rate: float = None, read_mode: str = None,
                     on_done: CallbackFnType = None,
                     on_error: CallbackFnType = None,
                     on_always: CallbackFnType = None,
                     **kwargs):
        """ Use a set of files as the query source for searching on the current Flow
        :param patterns: The pattern may contain simple shell-style wildcards, e.g. '\*.py', '[\*.zip, \*.gz]'
        :param recursive: If recursive is true, the pattern '**' will match any files and
                    zero or more directories and subdirectories.
        :param size: the maximum number of the files
        :param sampling_rate: the sampling rate between [0, 1]
        :param read_mode: specifies the mode in which the file
                is opened. 'r' for reading in text mode, 'rb' for reading in
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :return: results
        """
        from ...clients.sugary_io import _input_files
        return self._get_client(**kwargs).search(_input_files(patterns, recursive, size, sampling_rate, read_mode),
                                                 on_done, on_error, on_always, data_type=DataInputType.CONTENT,
                                                 **kwargs)

    @deprecated_alias(input_fn=('inputs', 0), buffer=('inputs', 1), callback=('on_done', 1), output_fn=('on_done', 1))
    def search_lines(self,
                     lines: Union[Iterable[str], TextIO] = None,
                     filepath: str = None,
                     field_resolver: Dict[str, str] = None,
                     size: int = None,
                     sampling_rate: float = None,
                     read_mode: str = 'r',
                     line_format: str = 'json',
                     on_done: CallbackFnType = None,
                     on_error: CallbackFnType = None,
                     on_always: CallbackFnType = None,
                     **kwargs):
        """ Use a list of files as the query source for searching on the current Flow
        :param filepath: a text file that each line contains a document
        :param lines: a list of strings, each is considered as d document
        :param size: the maximum number of the documents
        :param sampling_rate: the sampling rate between [0, 1]
        :param read_mode: specifies the mode in which the file
                is opened. 'r' for reading in text mode, 'rb' for reading in binary
        :param line_format: the format of each line ``json`` or ``csv``
        :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
            names defined in Protobuf. This is only used when the given ``document`` is
            a JSON string or a Python dict.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :return: results
        """
        from ...clients.sugary_io import _input_lines
        return self._get_client(**kwargs).search(
            _input_lines(lines, filepath,
                         size=size,
                         sampling_rate=sampling_rate,
                         read_mode=read_mode,
                         line_format=line_format,
                         field_resolver=field_resolver),
            on_done, on_error, on_always, data_type=DataInputType.AUTO, **kwargs)

    def search_ndjson(self,
                      lines: Union[Iterable[str], TextIO],
                      field_resolver: Dict[str, str] = None,
                      size: int = None,
                      sampling_rate: float = None,
                      on_done: CallbackFnType = None,
                      on_error: CallbackFnType = None,
                      on_always: CallbackFnType = None,
                      **kwargs):
        """ Use a list of files as the query source for searching on the current Flow
        :param lines: a list of strings, each is considered as d document
        :param size: the maximum number of the documents
        :param sampling_rate: the sampling rate between [0, 1]
        :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
            names defined in Protobuf. This is only used when the given ``document`` is
            a JSON string or a Python dict.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :return: results
        """
        from ...clients.sugary_io import _input_ndjson
        return self._get_client(**kwargs).search(
            _input_ndjson(lines,
                          size=size,
                          sampling_rate=sampling_rate,
                          field_resolver=field_resolver),
            on_done, on_error, on_always, data_type=DataInputType.AUTO, **kwargs)

    @deprecated_alias(input_fn=('inputs', 0), buffer=('inputs', 1), callback=('on_done', 1), output_fn=('on_done', 1))
    def index(self, inputs: InputType,
              on_done: CallbackFnType = None,
              on_error: CallbackFnType = None,
              on_always: CallbackFnType = None,
              **kwargs):
        """Do indexing on the current Flow
        :param inputs: An iterator of bytes. If not given, then you have to specify it in **kwargs**.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :return: results
        """
        return self._get_client(**kwargs).index(inputs, on_done, on_error, on_always, **kwargs)

    @deprecated_alias(input_fn=('inputs', 0))
    def update(self, inputs: InputType,
               on_done: CallbackFnType = None,
               on_error: CallbackFnType = None,
               on_always: CallbackFnType = None,
               **kwargs):
        """Updates Documents on the current Flow

        :param inputs: An iterator of bytes. If not given, then you have to specify it in **kwargs**.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        """
        self._get_client(**kwargs).update(inputs, on_done, on_error, on_always, **kwargs)

    def delete(self, ids: Iterable[str],
               on_done: CallbackFnType = None,
               on_error: CallbackFnType = None,
               on_always: CallbackFnType = None,
               **kwargs):
        """Do deletion on the current Flow

        :param ids: An iterator of bytes. If not given, then you have to specify it in **kwargs**.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        """
        self._get_client(**kwargs).delete(ids, on_done, on_error, on_always, **kwargs)

    @deprecated_alias(input_fn=('inputs', 0), buffer=('inputs', 1), callback=('on_done', 1), output_fn=('on_done', 1))
    def search(self, inputs: InputType,
               on_done: CallbackFnType = None,
               on_error: CallbackFnType = None,
               on_always: CallbackFnType = None,
               **kwargs):
        """Do searching on the current Flow
        It will start a :py:class:`CLIClient` and call :py:func:`search`.

        :param inputs: An iterator of bytes. If not given, then you have to specify it in **kwargs**.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :return: results
        """
        return self._get_client(**kwargs).search(inputs, on_done, on_error, on_always, **kwargs)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

if False:
    from argparse import Namespace


def pod(args: 'Namespace'):
    """Start a Pod"""
    from jina.peapods import Pod
    try:
        with Pod(args) as p:
            p.join()
    except KeyboardInterrupt:
        pass


def pea(args: 'Namespace'):
    """Start a Pea"""
    from jina.peapods import Pea
    try:
        with Pea(args) as p:
            p.join()
    except KeyboardInterrupt:
        pass


def gateway(args: 'Namespace'):
    """Start a Gateway Pod"""
    pod(args)


def check(args: 'Namespace'):
    """Check jina config, settings, imports, network etc"""
    from jina.checker import ImportChecker
    ImportChecker(args)


def ping(args: 'Namespace'):
    from jina.checker import NetworkChecker
    NetworkChecker(args)


def client(args: 'Namespace'):
    """Start a client connects to the gateway"""
    from jina.clients import Client
    Client(args)


def export_api(args: 'Namespace'):
    import json
    from .export import api_to_dict
    from jina.jaml import JAML
    from jina import __version__
    from jina.logging import default_logger
    from jina.schemas import get_full_schema

    if args.yaml_path:
        dump_api = api_to_dict()
        for yp in args.yaml_path:
            f_name = (yp % __version__) if '%s' in yp else yp
            with open(f_name, 'w', encoding='utf8') as fp:
                JAML.dump(dump_api, fp)
            default_logger.info(f'API is exported to {f_name}')

    if args.json_path:
        dump_api = api_to_dict()
        for jp in args.json_path:
            f_name = (jp % __version__) if '%s' in jp else jp
            with open(f_name, 'w', encoding='utf8') as fp:
                json.dump(dump_api, fp, sort_keys=True)
            default_logger.info(f'API is exported to {f_name}')

    if args.schema_path:
        dump_api = get_full_schema()
        for jp in args.schema_path:
            f_name = (jp % __version__) if '%s' in jp else jp
            with open(f_name, 'w', encoding='utf8') as fp:
                json.dump(dump_api, fp, sort_keys=True)
            default_logger.info(f'API is exported to {f_name}')


def hello_world(args: 'Namespace'):
    from jina.helloworld.fashion import hello_world
    hello_world(args)


def hello(args: 'Namespace'):
    if args.hello == 'fashion':
        from jina.helloworld.fashion import hello_world
    elif args.hello == 'chatbot':
        from jina.helloworld.chatbot import hello_world
    elif args.hello == 'multimodal':
        from jina.helloworld.multimodal import hello_world
    else:
        raise ValueError(f'must be one of [`fashion`, `chatbot`, `multimodal`]')

    hello_world(args)


def flow(args: 'Namespace'):
    """Start a Flow from a YAML file or a docker image"""
    from jina.flow import Flow
    if args.uses:
        f = Flow.load_config(args.uses)
        with f:
            f.block()
    else:
        from jina.logging import default_logger
        default_logger.critical('start a flow from CLI requires a valid "--uses"')


def optimizer(args: 'Namespace'):
    """Start an optimization from a YAML file"""
    from jina.optimizers import run_optimizer_cli
    run_optimizer_cli(args)


def hub(args: 'Namespace'):
    """Start a hub builder for build, push, pull"""
    from jina.docker.hubio import HubIO
    getattr(HubIO(args), args.hub)()



"""Argparser module for ZED runtime"""
import argparse

from ...helper import add_arg_group, _SHOW_ALL_ARGS
from .... import __default_host__
from ....enums import OnErrorStrategy, SocketType
from ....helper import random_port


def mixin_zed_runtime_parser(parser):
    """Mixing in arguments required by :class:`ZEDRuntime` into the given parser.
    :param parser: the parser instance to which we add arguments
    """

    gp = add_arg_group(parser, title='ZEDRuntime')

    gp.add_argument('--uses', type=str, default='_pass',
                    help='''
The config of the executor, it could be one of the followings: 
- an Executor-level YAML file path (.yml, .yaml, .jaml) 
- a name of a class inherited from `jina.Executor`
- a docker image (must start with `docker://`)
- builtin executors, e.g. `_pass`, `_logforward`, `_merge` 
- the string literal of a YAML config (must start with `!`)
- the string literal of a JSON config
- the string literal of a YAML driver config (must start with `- !!`)

When use it under Python, one can use the following values additionally:
- a Python dict that represents the config
- a text file stream has `.read()` interface
''')
    gp.add_argument('--py-modules', type=str, nargs='*', metavar='PATH',
                    help='''
The customized python modules need to be imported before loading the executor

Note, when importing multiple files and there is a dependency between them, then one has to write the dependencies in 
reverse order. That is, if `__init__.py` depends on `A.py`, which again depends on `B.py`, then you need to write: 

--py-modules __init__.py --py-modules B.py --py-modules A.py

''')

    gp.add_argument('--port-in', type=int, default=random_port(),
                    help='The port for input data, default a random port between [49152, 65535]')
    gp.add_argument('--port-out', type=int, default=random_port(),
                    help='The port for output data, default a random port between [49152, 65535]')
    gp.add_argument('--host-in', type=str, default=__default_host__,
                    help=f'The host address for input, by default it is {__default_host__}')
    gp.add_argument('--host-out', type=str, default=__default_host__,
                    help=f'The host address for output, by default it is {__default_host__}')
    gp.add_argument('--socket-in', type=SocketType.from_string, choices=list(SocketType),
                    default=SocketType.PULL_BIND,
                    help='The socket type for input port')
    gp.add_argument('--socket-out', type=SocketType.from_string, choices=list(SocketType),
                    default=SocketType.PUSH_BIND,
                    help='The socket type for output port')

    gp.add_argument('--dump-interval', type=int, default=240,
                    help='Serialize the model in the pod every n seconds if model changes. '
                         '-1 means --read-only. ')
    gp.add_argument('--read-only', action='store_true', default=False,
                    help='If set, do not allow the pod to modify the model, '
                         'dump_interval will be ignored')

    gp.add_argument('--memory-hwm', type=int, default=-1,
                    help='The memory high watermark of this pod in Gigabytes, pod will restart when this is reached. '
                         '-1 means no restriction')

    gp.add_argument('--on-error-strategy', type=OnErrorStrategy.from_string, choices=list(OnErrorStrategy),
                    default=OnErrorStrategy.IGNORE,
                    help='''
The skip strategy on exceptions.

- IGNORE: Ignore it, keep running all Drivers & Executors logics in the sequel flow
- SKIP_EXECUTOR: Skip all Executors in the sequel, but drivers are still called
- SKIP_HANDLE: Skip all Drivers & Executors in the sequel, only `pre_hook` and `post_hook` are called
- THROW_EARLY: Immediately throw the exception, the sequel flow will not be running at all 
                    
Note, `IGNORE`, `SKIP_EXECUTOR` and `SKIP_HANDLE` do not guarantee the success execution in the sequel flow. If something 
is wrong in the upstream, it is hard to carry this exception and moving forward without any side-effect.
''')

    gp.add_argument('--num-part', type=int, default=0,
                    help='the number of messages expected from upstream, 0 and 1 means single part'
                    if _SHOW_ALL_ARGS else argparse.SUPPRESS)



from cli.export import api_to_dict

from .helper import _cli_to_schema

schema_pod = _cli_to_schema(
    api_to_dict(),
    'pod',
    extras={
        'needs': {
            'description': 'The name of the Pod(s) that this Pod receives data from. '
                           'One can also use `gateway` to '
                           'indicate the connection with the gateway.\n'
                           'This is useful to create parallelization. '
                           'By default the Flow always works sequentially '
                           'following the defined order of the Pods.',
            'type': ['array', 'string'],
            'items': {
                'type': 'string',
                'minItems': 1,
                "uniqueItems": True
            }
        },
        'method': {
            'description': 'The method to use when appending the Pod to the Flow',
            'type': 'string',
            'enum': ['add', 'needs', 'inspect', 'needs_all',
                     'gather_inspect'],
            'default': 'add'
        },
    },
    allow_addition=False,
    description='Define the config of a Pod.')



from jina.executors.encoders import BaseTextEncoder
from jina.executors.decorators import batching, as_ndarray
import numpy as np

class OneHotTextEncoder(BaseTextEncoder):
    """
    Encode the characters into one-hot vectors.

    ONLY FOR TESTING USAGES.

    :param on_value: the default value for the locations
        represented by characters
    :param off_value: the default value for the locations
        not represented by characters
    """

    def __init__(self,
                 on_value: float = 1,
                 off_value: float = 0,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.offset = 32
        self.dim = 127 - self.offset + 2  # only the Unicode code point between 32 and 127 are embedded, and the rest are considered as ``UNK```
        self.unk = self.dim
        self.on_value = on_value
        self.off_value = off_value
        self.embeddings = None

    def post_init(self):
        """Set embeddings"""
        self.embeddings = np.eye(self.dim) * self.on_value + \
                          (np.ones((self.dim, self.dim)) - np.eye(self.dim)) * self.off_value


    @batching
    @as_ndarray
    def encode(self, data: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
        Encode data into one-hot vectors.

        :param data: each row is one character,
            an 1d array of string type (data.dtype.kind == 'U') in size B
        :return: an ndarray of `B x D`. Where B is the `Batch size`
            and `D` the dimension.
        """
        output = []
        for r in data:
            r_emb = [ord(c) - self.offset if self.offset <= ord(c) <= 127 else self.unk for c in r]
            output.append(self.embeddings[r_emb, :].sum(axis=0))
        return np.array(output)


from typing import Sequence, Any
import numpy as np
from jina.executors.evaluators.rank import BaseRankingEvaluator


class AveragePrecisionEvaluator(BaseRankingEvaluator):
    """
    Evaluates the Average Precision of the search.
    https://en.wikipedia.org/wiki/Evaluation_measures_(information_retrieval)#Average_precision

    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def evaluate(self, actual: Sequence[Any], desired: Sequence[Any], *args, **kwargs) -> float:
        """"
        Evaluate the Average Precision of the search.

        :param actual: the matched document identifiers from the request
            as matched by Indexers and Rankers
        :param desired: A list of all the relevant IDs. All documents
            identified in this list are considered to be relevant
        :return: the evaluation metric value for the request document
        :param args:  Additional positional arguments
        :param kwargs: Additional keyword arguments
        """

        if len(desired) == 0 or len(actual) == 0:
            return 0.0

        desired_set = set(desired)

        def _precision(eval_at: int):
            if actual[eval_at - 1] not in desired_set:
                return 0.0
            actual_at_k = actual[:eval_at]
            ret = len(set(actual_at_k).intersection(desired_set))
            sub = len(actual_at_k)
            return ret / sub if sub != 0 else 0.

        precisions = list(map(lambda eval_at: _precision(eval_at), range(1, len(actual) + 1)))
        return sum(precisions) / len(desired)



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Optional
import numpy as np

from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders.frameworks import BaseTFEncoder


class ImageKerasEncoder(BaseTFEncoder):
    """
    :class:`ImageKerasEncoder` encodes data from a ndarray,
    potentially B x (Channel x Height x Width) into a ndarray of `B x D`.

    Where `B` is the batch size and `D` is the Dimension.

    The :class:`ImageKerasEncoder` wraps the models from
    `tensorflow.keras.applications`. <https://keras.io/applications/>`_.

    :param model_name: the name of the model. Supported models include
        ``DenseNet121``, ``DenseNet169``, ``DenseNet201``,
        ``InceptionResNetV2``, ``InceptionV3``, ``MobileNet``,
        ``MobileNetV2``, ``NASNetLarge``, ``NASNetMobile``,
        ``ResNet101``, ``ResNet152``, ``ResNet50``, ``ResNet101V2``,
        ``ResNet152V2``, ``ResNet50V2``, ``VGG16``, ``VGG19``,
        ``Xception``,
    :param img_shape: The shape of the image to be encoded.
    :param pool_strategy: the pooling strategy. Options are:
        - `None`: Means that the output of the model will be the 4D tensor
            output of the last convolutional block.
        - `avg`: ;eans that global average pooling will be applied to the
            output of the last convolutional block, and thus the output of
            the model will be a 2D tensor.
        - `max`: Means that global max pooling will be applied.

    :param channel_axis: the axis id of the channel, -1 indicate the color channel
        info at the last axis. If given other, then ``np.moveaxis(data, channel_axis, -1)``
        is performed before :meth:`encode`.
    :param args: additional positional arguments.
    :param kwargs: additional positional arguments.
    """

    def __init__(self,
                 model_name: Optional[str] = None,
                 img_shape: int = 96,
                 pool_strategy: str = 'avg',
                 channel_axis: int = 1,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.model_name = model_name or 'MobileNetV2'
        self.pool_strategy = pool_strategy
        self.img_shape = img_shape
        self.channel_axis = channel_axis

    def post_init(self):
        """Load model and set its attributes"""
        super().post_init()
        self.to_device()
        import tensorflow as tf
        model = getattr(tf.keras.applications, self.model_name)(
            input_shape=(self.img_shape, self.img_shape, 3),
            include_top=False,
            pooling=self.pool_strategy,
            weights='imagenet')
        model.trainable = False
        self.model = model

    @batching
    @as_ndarray
    def encode(self, data: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
        Encode data into a ndarray of `B x D`. `
        B` is the batch size and `D` is the Dimension.

        :param data: Image to be encoded, expected a `np.ndarray` of
            BatchSize x (Channel x Height x Width).
        :param args: additional positional arguments.
        :param kwargs: additional positional arguments.
        :return: Encoded result as a `BatchSize x D` numpy ``ndarray``,
            `D` is the output dimension
        """
        if self.channel_axis != -1:
            data = np.moveaxis(data, self.channel_axis, -1)
        return self.model(data)



"""Argparser module for container runtimes"""
from ...helper import add_arg_group, KVAppendAction


def mixin_container_runtime_parser(parser):
    """Mixing in arguments required by :class:`ContainerRuntime` into the given parser.
    :param parser: the parser instance to which we add arguments
    """
    gp = add_arg_group(parser, title='ContainerRuntime')

    gp.add_argument('--uses-internal', type=str, default='BaseExecutor',
                    help='''
The config runs inside the Docker container. 

Syntax and function are the same as `--uses`. This is designed when `--uses="docker://..."` this config is passed to 
the Docker container.
''')
    gp.add_argument('--entrypoint', type=str,
                    help='The entrypoint command overrides the ENTRYPOINT in Docker image. '
                         'when not set then the Docker image ENTRYPOINT takes effective.')
    gp.add_argument('--docker-kwargs', action=KVAppendAction,
                    metavar='KEY=VALUE', nargs='*',
                    help='''
Dictionary of kwargs arguments that will be passed to Docker SDK when starting the docker '
container. 

More details can be found in the Docker SDK docs:  https://docker-py.readthedocs.io/en/stable/

''')
    gp.add_argument('--pull-latest', action='store_true', default=False,
                    help='Pull the latest image before running')
    gp.add_argument('--volumes', type=str, nargs='*', metavar='DIR',
                    help='''
The path on the host to be mounted inside the container. 

Note, 
- If separated by `:`, then the first part will be considered as the local host path and the second part is the path in the container system. 
- If no split provided, then the basename of that directory will be mounted into container's root path, e.g. `--volumes="/user/test/my-workspace"` will be mounted into `/my-workspace` inside the container. 
- All volumes are mounted with read-write mode.
''')



import numpy as np

from ..embedding import BaseEmbeddingEvaluator, expand_vector


class EuclideanEvaluator(BaseEmbeddingEvaluator):
    """A :class:`EuclideanEvaluator` evaluates the distance between actual and desired embeddings computing
    the euclidean distance between them
    """

    metric = 'EuclideanDistance'

    def evaluate(self, actual: 'np.array', desired: 'np.array', *args, **kwargs) -> float:
        """"
        :param actual: the embedding of the document (resulting from an Encoder)
        :param desired: the expected embedding of the document
        :return the evaluation metric value for the request document
        """
        actual = expand_vector(actual)
        desired = expand_vector(desired)
        return _euclidean(_ext_A(actual), _ext_B(desired))


# duplicate on purpose, to be migrated to the Hub
def _get_ones(x, y):
    return np.ones((x, y))


def _ext_A(A):
    nA, dim = A.shape
    A_ext = _get_ones(nA, dim * 3)
    A_ext[:, dim:2 * dim] = A
    A_ext[:, 2 * dim:] = A ** 2
    return A_ext


def _ext_B(B):
    nB, dim = B.shape
    B_ext = _get_ones(dim * 3, nB)
    B_ext[:dim] = (B ** 2).T
    B_ext[dim:2 * dim] = -2.0 * B.T
    del B
    return B_ext


def _euclidean(A_ext, B_ext):
    sqdist = A_ext.dot(B_ext).clip(min=0)
    return np.sqrt(sqdist)


def _norm(A):
    return A / np.linalg.norm(A, ord=2, axis=1, keepdims=True)


def _cosine(A_norm_ext, B_norm_ext):
    return A_norm_ext.dot(B_norm_ext).clip(min=0) / 2


"""Argparser module for hub push & pull"""
from ..helper import add_arg_group


def mixin_hub_pushpull_parser(parser):
    """Add the arguments for hub push pull to the parser

    :param parser: the parser configure
    """
    gp = add_arg_group(parser, title='Push/Pull')
    gp.add_argument('name', type=str, help='The name of the image.')
    gp.add_argument('--no-overwrite', action='store_true', default=False,
                    help='If set, do not allow overwriting existing images (based on module version and jina version)')



"""Argparser module for the export API"""
from .base import set_base_parser


def set_export_api_parser(parser=None):
    """Set the parser for the API export

    :param parser: an optional existing parser to build upon
    :return: the parser
    """
    if not parser:
        parser = set_base_parser()

    parser.add_argument('--yaml-path', type=str, nargs='*', metavar='PATH',
                        help='The YAML file path for storing the exported API')
    parser.add_argument('--json-path', type=str, nargs='*', metavar='PATH',
                        help='The JSON file path for storing the exported API')
    parser.add_argument('--schema-path', type=str, nargs='*', metavar='PATH',
                        help='The JSONSchema file path for storing the exported API')
    return parser



import multiprocessing
import threading
from multiprocessing.synchronize import Event

from ...enums import RuntimeBackendType


def _get_event(obj) -> Event:
    if isinstance(obj, threading.Thread):
        return threading.Event()
    elif isinstance(obj, multiprocessing.Process):
        return multiprocessing.Event()
    else:
        raise NotImplementedError


def _make_or_event(obj, *events) -> Event:
    or_event = _get_event(obj)

    def or_set(self):
        self._set()
        self.changed()

    def or_clear(self):
        self._clear()
        self.changed()

    def orify(e, changed_callback):
        e._set = e.set
        e._clear = e.clear
        e.changed = changed_callback
        e.set = lambda: or_set(e)
        e.clear = lambda: or_clear(e)

    def changed():
        bools = [e.is_set() for e in events]
        if any(bools):
            or_event.set()
        else:
            or_event.clear()

    for e in events:
        orify(e, changed)
    changed()
    return or_event


class PeaType(type):
    """Type of :class:`Pea`, metaclass of :class:`BasePea`."""
    _dct = {}

    def __new__(cls, name, bases, dct):
        """
        Create and register a new class with this meta class.

        :param name: name of the :class:`Pea`
        :param bases: bases of :class:`Pea`
        :param dct: arguments dictionary
        :return: registered class
        """
        _cls = super().__new__(cls, name, bases, dct)
        PeaType._dct.update({name: {'cls': cls,
                                    'name': name,
                                    'bases': bases,
                                    'dct': dct}})
        return _cls

    def __call__(cls, *args, **kwargs) -> 'PeaType':
        """
        change runtime backend

        :param args: arguments
        :param kwargs: keyword arguments
        :return: Call self as a function.
        """
        # switch to the new backend
        _cls = {
            RuntimeBackendType.THREAD: threading.Thread,
            RuntimeBackendType.PROCESS: multiprocessing.Process,
        }.get(getattr(args[0], 'runtime_backend', RuntimeBackendType.THREAD))

        # rebuild the class according to mro
        for c in cls.mro()[-2::-1]:
            arg_cls = PeaType._dct[c.__name__]['cls']
            arg_name = PeaType._dct[c.__name__]['name']
            arg_dct = PeaType._dct[c.__name__]['dct']
            _cls = super().__new__(arg_cls, arg_name, (_cls,), arg_dct)

        return type.__call__(_cls, *args, **kwargs)
